import pytest

from sdgx.data_models.inspectors.manager import InspectorManager


@pytest.fixture
def manager():
    yield InspectorManager()


@pytest.mark.parametrize(
    "supported_inspector",
    [
        "DiscreteInspector",
    ],
)
def test_manager(supported_inspector, manager: InspectorManager):
    assert manager._normalize_name(supported_inspector) in manager.registed_inspectors


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from sdgx.data_processors.base import DataProcessor


class Formatter(DataProcessor):
    """
    Base class for formatters.

    Formatter is used to convert data from one format to another.

    For example, parse datetime into timestamp when trainning,
    and format timestamp into datetime when sampling.

    Difference with :ref:`Transformer`:
    - When a single column is used as input, use formatter for formatting issues.
    - When a whole table is used as input, use :ref:`Transformer`.
    - :ref:`Transformer` sometimes implements some functions with the help of Formatter.

    """



import pytest

from sdgx.models.ml.single_table.ctgan import CTGANSynthesizerModel


@pytest.fixture
def ctgan():
    yield CTGANSynthesizerModel(epochs=1)


@pytest.fixture
def save_model_path(tmp_path):
    filename = tmp_path / "ctgan-model.pkl"
    yield filename
    filename.unlink(missing_ok=True)


def test_ctgan(
    ctgan: CTGANSynthesizerModel,
    demo_single_table_metadata,
    demo_single_table_data_loader,
    save_model_path,
):
    ctgan.fit(demo_single_table_metadata, demo_single_table_data_loader)
    ctgan.sample(10)
    ctgan.save(save_model_path)
    model = CTGANSynthesizerModel.load(save_model_path)
    model.sample(10)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

import pytest

from sdgx.data_loader import DataLoader


def test_dataloader(demo_single_table_data_loader: DataLoader):
    assert len(demo_single_table_data_loader) == 48843
    assert (
        sorted(demo_single_table_data_loader.columns())
        == sorted(demo_single_table_data_loader.keys())
        == sorted(
            [
                "age",
                "workclass",
                "fnlwgt",
                "education",
                "education-num",
                "marital-status",
                "occupation",
                "relationship",
                "race",
                "sex",
                "capitalgain",
                "capitalloss",
                "hoursperweek",
                "native-country",
                "class",
            ]
        )
    )
    assert demo_single_table_data_loader.shape == (48843, 15)
    assert demo_single_table_data_loader.load_all().shape == (48843, 15)

    assert demo_single_table_data_loader[:100].shape == (100, 15)
    assert demo_single_table_data_loader[100:].shape == (48843 - 100, 15)
    assert demo_single_table_data_loader[100:10000].shape == (10000 - 100, 15)
    assert demo_single_table_data_loader[100:10000:2].shape == ((10000 - 100) // 2, 15)

    assert demo_single_table_data_loader[["age", "workclass"]].shape == (48843, 2)

    for df in demo_single_table_data_loader.iter():
        assert len(df) == demo_single_table_data_loader.chunksize
        break


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



class SdgxError(Exception):
    """
    Base class for exceptions in this module.
    """


class NonParametricError(Exception):
    """
    Exception to indicate that a model is not parametric.
    """


class ManagerError(SdgxError):
    """
    Exception to indicate that exception when using manager.
    """


class NotFoundError(ManagerError):
    """
    Exception to indicate that a model is not found.
    """


class RegisterError(ManagerError):
    """
    Exception to indicate that exception when registering.
    """


class InitializationError(ManagerError):
    """
    Exception to indicate that exception when initializing model.
    """


class ManagerLoadModelError(ManagerError):
    """
    Exception to indicate that exception when loading model for :ref:`ModelManager`.
    """


class SynthesizerInitError(ManagerError):
    """
    Exception to indicate that exception when synthesizing model.
    """


class CacheError(SdgxError):
    """
    Exception to indicate that exception when using cache.
    """


class MetadataInitError(SdgxError):
    """
    Exception to indicate that exception when initializing metadata.
    """


class DataLoaderInitError(SdgxError):
    """
    Exception to indicate that exception when initializing dataloader.
    """



# # 本例用于测试 transformer 对象的正常使用

# import os
# import shutil
# import sys

# import pytest

# _HERE = os.path.dirname(__file__)
# sys.path.append(os.getcwd())

# from sdgx.utils.io.csv_utils import *


# def test_csv_tools():
#     pass


# if __name__ == "__main__":
#     pytest.main(["-vv", "-s", __file__])



# import os

# import pytest

# _HERE = os.path.dirname(__file__)

# import sys

# sys.path.append(os.getcwd())


# from sdgx.models.single_table.ctgan import CTGAN
# from sdgx.utils.io.csv_utils import *


# def test_helloworld():
#     # 读取数据
#     demo_data, discrete_cols = get_demo_single_table()
#     model = CTGAN(epochs=1)
#     model.fit(demo_data, discrete_cols)

#     # 生成合成数据
#     sampled_data = model.sample(1000)


# if __name__ == "__main__":
#     pytest.main(["-vv", "-s", __file__])



import pytest

from sdgx.data_connectors.csv_connector import CsvConnector
from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata


@pytest.fixture
def dataloader(demo_single_table_path, cacher_kwargs):
    d = DataLoader(CsvConnector(path=demo_single_table_path), cacher_kwargs=cacher_kwargs)
    yield d
    d.finalize(clear_cache=True)


@pytest.fixture
def metadata(dataloader):
    yield Metadata.from_dataloader(dataloader)


def test_metadata(metadata: Metadata):
    assert metadata.discrete_columns == metadata.get("discrete_columns")
    assert metadata.model_dump_json()


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pytest

from sdgx.cachers.manager import CacherManager


@pytest.fixture
def manager():
    yield CacherManager()


@pytest.mark.parametrize(
    "supported_data_connector",
    ["NoCache", "MemoryCache", "DiskCache"],
)
def test_manager(supported_data_connector, manager: CacherManager):
    assert manager._normalize_name(supported_data_connector) in manager.registed_cachers


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from pathlib import Path

import pandas as pd

from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata


class SynthesizerModel:
    def fit(self, metadata: Metadata, dataloader: DataLoader, *args, **kwargs):
        raise NotImplementedError

    def sample(self, count: int, *args, **kwargs) -> pd.DataFrame:
        raise NotImplementedError

    def save(self, path: str | Path):
        raise NotImplementedError

    @classmethod
    def load(cls, path: str | Path) -> "SynthesizerModel":
        raise NotImplementedError



from __future__ import annotations

import json
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List

import pandas as pd
from pydantic import BaseModel

from sdgx.data_loader import DataLoader
from sdgx.data_models.inspectors.manager import InspectorManager
from sdgx.exceptions import MetadataInitError
from sdgx.utils import cache

# TODO: Design metadata for relationships...
# class DType(Enum):
#     datetime = "datetime"
#     timestamp = "timestamp"
#     numeric = "numeric"
#     category = "category"


# class Relationship:
#     pass


class Metadata(BaseModel):
    discrete_columns: List[str] = []
    _extend: Dict[str, Any] = {}

    def get(self, key: str, default=None) -> Any:
        return getattr(self, key, getattr(self._extend, key, default))

    def set(self, key: str, value: Any):
        if key == "_extend":
            raise MetadataInitError("Cannot set _extend directly")

        if key in self.model_fields:
            setattr(self, key, value)
        else:
            self._extend[key] = value

    def update(self, attributes: dict[str, Any]):
        for k, v in attributes.items():
            self.set(k, v)

        return self

    @classmethod
    def from_dataloader(
        cls,
        dataloader: DataLoader,
        max_chunk: int = 10,
        include_inspectors: list[str] | None = None,
        exclude_inspectors: list[str] | None = None,
        inspector_init_kwargs: dict[str, Any] | None = None,
    ) -> "Metadata":
        inspectors = InspectorManager().init_inspcetors(
            include_inspectors, exclude_inspectors, **(inspector_init_kwargs or {})
        )
        for i, chunk in enumerate(dataloader.iter()):
            for inspector in inspectors:
                inspector.fit(chunk)
            if all(i.ready for i in inspectors) or i > max_chunk:
                break

        metadata = Metadata()
        for inspector in inspectors:
            metadata.update(inspector.inspect())

        return metadata

    @classmethod
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        include_inspectors: list[str] | None = None,
        exclude_inspectors: list[str] | None = None,
        inspector_init_kwargs: dict[str, Any] | None = None,
    ) -> "Metadata":
        inspectors = InspectorManager().init_inspcetors(
            include_inspectors, exclude_inspectors, **(inspector_init_kwargs or {})
        )
        for inspector in inspectors:
            inspector.fit(df)

        metadata = Metadata()
        for inspector in inspectors:
            metadata.update(inspector.inspect())

        return metadata

    def save(self, path: str | Path):
        with path.open("w") as f:
            f.write(self.model_dump_json())

    @classmethod
    def load(cls, path: str | Path) -> "Metadata":
        path = Path(path).expanduser().resolve()
        attributes = json.load(path.open("r"))
        return Metadata().update(attributes)



from sdgx.data_connectors.csv_connector import CsvConnector
from sdgx.models.ml.single_table.ctgan import CTGANSynthesizerModel
from sdgx.synthesizer import Synthesizer
from sdgx.utils import download_demo_data

dataset_csv = download_demo_data()
data_connector = CsvConnector(path=dataset_csv)
synthesizer = Synthesizer(
    model=CTGANSynthesizerModel(epochs=1),  # For quick demo
    data_connector=data_connector,
)
synthesizer.fit()
sampled_data = synthesizer.sample(1000)
synthesizer.cleanup()  # Clean all cache


from sdgx.metrics.column.jsd import JSD

JSD = JSD()


selected_columns = ["workclass"]
isDiscrete = True
metrics = JSD.calculate(data_connector.read(), sampled_data, selected_columns, isDiscrete)

print("JSD metric of column %s: %g" % (selected_columns[0], metrics))



from __future__ import annotations

import threading
import urllib.request
from pathlib import Path

import pandas as pd

from sdgx.log import logger

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

__all__ = ["download_demo_data", "get_demo_single_table", "cache", "Singleton"]


def download_demo_data(data_dir: str | Path = "./dataset") -> Path:
    """
    Download demo data if not exist

    Args:
        data_dir(str | Path): data directory

    Returns:
        pathlib.Path: demo data path
    """
    data_dir = Path(data_dir).expanduser().resolve()
    demo_data_path = data_dir / "adult.csv"
    if not demo_data_path.exists():
        # Download from datahub
        demo_data_path.parent.mkdir(parents=True, exist_ok=True)

        logger.info("Downloading demo data from datahub.io to {}".format(demo_data_path))
        url = "https://datahub.io/machine-learning/adult/r/adult.csv"
        urllib.request.urlretrieve(url, demo_data_path)

    return demo_data_path


def get_demo_single_table(data_dir: str | Path = "./dataset"):
    """
    Get demo single table as DataFrame and discrete columns names

    Args:
        data_dir(str | Path): data directory

    Returns:

        pd.DataFrame: demo single table
        list: discrete columns
    """
    demo_data_path = download_demo_data(data_dir)
    pd_obj = pd.read_csv(demo_data_path)
    discrete_cols = [
        "workclass",
        "education",
        "marital-status",
        "occupation",
        "relationship",
        "race",
        "sex",
        "native-country",
        "class",
    ]
    return pd_obj, discrete_cols


class Singleton(type):
    """
    metaclass for singleton, thread-safe.
    """

    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



from __future__ import annotations

from pathlib import Path
from typing import Any, Generator

import pandas as pd

from sdgx.data_connectors.base import DataConnector
from sdgx.data_connectors.generator_connector import GeneratorConnector
from sdgx.data_connectors.manager import DataConnectorManager
from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.base import DataProcessor
from sdgx.data_processors.manager import DataProcessorManager
from sdgx.exceptions import SynthesizerInitError
from sdgx.log import logger
from sdgx.models.base import SynthesizerModel
from sdgx.models.manager import ModelManager


class Synthesizer:
    """
    Synthesizer takes all logic of preprocessing data, fit model and sample
    """

    METADATA_SAVE_NAME = "metadata.json"
    MODEL_SAVE_NAME = "model.pkl"

    def __init__(
        self,
        model: str | SynthesizerModel | type[SynthesizerModel],
        model_path: None | str | Path = None,
        metadata: None | Metadata = None,
        metadata_path: None | str | Path = None,
        data_connector: None | str | DataConnector = None,
        data_connectors_kwargs: None | dict[str, Any] = None,
        raw_data_loaders_kwargs: None | dict[str, Any] = None,
        processored_data_loaders_kwargs: None | dict[str, Any] = None,
        data_processors: None | list[str | DataProcessor] = None,
        data_processors_kwargs: None | dict[str, dict[str, Any]] = None,
        model_kwargs: None | dict[str, Any] = None,
    ):
        # Init data connectors
        if isinstance(data_connector, str):
            data_connector = DataConnectorManager().init_data_connector(
                data_connector, **(data_connectors_kwargs or {})
            )
        if data_connector:
            self.dataloader = DataLoader(
                data_connector,
                **(raw_data_loaders_kwargs or {}),
            )
        else:
            logger.warning("No data_connector provided, will not support `fit`")
            self.dataloader = None

        # Init data processors
        if not data_processors:
            data_processors = []
        self.data_processors_manager = DataProcessorManager()
        self.data_processors = [
            d
            if isinstance(d, DataProcessor)
            else self.data_processors_manager.init_data_processor(
                d, **(data_processors_kwargs or {})
            )
            for d in data_processors
        ]
        if metadata and metadata_path:
            raise SynthesizerInitError(
                "metadata and metadata_path cannot be specified at the same time"
            )

        # Load metadata
        if metadata:
            self.metadata = metadata
        elif metadata_path:
            self.metadata = Metadata.load(metadata_path)
        else:
            self.metadata = None

        # Init model
        self.model_manager = ModelManager()
        if isinstance(model, SynthesizerModel) and model_path:
            raise SynthesizerInitError(
                "model as instance and model_path cannot be specified at the same time"
            )
        if isinstance(model, str):
            self.model = self.model_manager.init_model(model, **(model_kwargs or {}))
        elif isinstance(model, SynthesizerModel):
            self.model = model
        elif model_path:
            self.model_manager.load(model, model_path)
        else:
            raise SynthesizerInitError("model or model_path must be specified")

        # Other arguments
        self.processored_data_loaders_kwargs = processored_data_loaders_kwargs or {}

    def save(self, save_dir: str | Path) -> Path:
        """
        Dump metadata and model to file
        """
        save_dir = Path(save_dir).expanduser().resolve()
        save_dir.mkdir(parents=True, exist_ok=True)
        if self.metadata:
            self.metadata.save(save_dir / self.METADATA_SAVE_NAME)
        self.model.save(save_dir / self.MODEL_SAVE_NAME)
        return save_dir

    @classmethod
    def load(
        cls,
        load_dir: str | Path,
        model: str | type[SynthesizerModel],
        metadata: None | Metadata = None,
        data_connector: None | str | DataConnector = None,
        data_connectors_kwargs: None | dict[str, Any] = None,
        raw_data_loaders_kwargs: None | dict[str, Any] = None,
        processored_data_loaders_kwargs: None | dict[str, Any] = None,
        data_processors: None | list[str | DataProcessor] = None,
        data_processors_kwargs: None | dict[str, dict[str, Any]] = None,
    ) -> "Synthesizer":
        """
        Load metadata and model from file
        """

        load_dir = Path(load_dir).expanduser().resolve()

        if not load_dir.exists():
            raise SynthesizerInitError(f"{load_dir.as_posix()} does not exist")
        model_path = load_dir / cls.MODEL_SAVE_NAME
        if not model_path.exists():
            raise SynthesizerInitError(
                f"{model_path.as_posix()} does not exist, cannot load model."
            )

        metadata_path = load_dir / cls.METADATA_SAVE_NAME
        if not metadata_path.exists():
            metadata_path = None

        return Synthesizer(
            model=model,
            model_path=load_dir / cls.MODEL_SAVE_NAME,
            metadata=metadata,
            metadata_path=metadata_path,
            data_connector=data_connector,
            data_connectors_kwargs=data_connectors_kwargs,
            raw_data_loaders_kwargs=raw_data_loaders_kwargs,
            processored_data_loaders_kwargs=processored_data_loaders_kwargs,
            data_processors=data_processors,
            data_processors_kwargs=data_processors_kwargs,
        )

    def fit(
        self,
        metadata: None | Metadata = None,
        inspector_max_chunk: int = 10,
        metadata_include_inspectors: None | list[str] = None,
        metadata_exclude_inspectors: None | list[str] = None,
        inspector_init_kwargs: None | dict[str, Any] = None,
        model_fit_kwargs: None | dict[str, Any] = None,
    ):
        if self.dataloader is None:
            raise SynthesizerInitError(
                "Cannot fit without dataloader, check `data_connector` parameter when initializing Synthesizer"
            )

        metadata = (
            metadata
            or self.metadata
            or Metadata.from_dataloader(
                self.dataloader,
                max_chunk=inspector_max_chunk,
                include_inspectors=metadata_include_inspectors,
                exclude_inspectors=metadata_exclude_inspectors,
                inspector_init_kwargs=inspector_init_kwargs,
            )
        )
        for d in self.data_processors:
            d.fit(metadata)

        def chunk_generator() -> Generator[pd.DataFrame, None, None]:
            for chunk in self.dataloader.iter():
                for d in self.data_processors:
                    chunk = d.convert(chunk)
                yield chunk

        processed_dataloader = DataLoader(
            GeneratorConnector(chunk_generator),
            **self.processored_data_loaders_kwargs,
        )
        try:
            self.model.fit(metadata, processed_dataloader, **(model_fit_kwargs or {}))
        finally:
            processed_dataloader.finalize(clear_cache=True)

    def sample(
        self,
        count: int,
        chunksize: None | int = None,
        metadata: None | Metadata = None,
        model_fit_kwargs: None | dict[str, Any] = None,
    ) -> pd.DataFrame | Generator[pd.DataFrame, None, None]:
        metadata = metadata or self.metadata
        if metadata:
            for d in self.data_processors:
                d.fit(metadata)
        if not model_fit_kwargs:
            model_fit_kwargs = {}

        if chunksize is None:
            sample_data = self.model.sample(count, **model_fit_kwargs)
            for d in self.data_processors:
                sample_data = d.reverse_convert(sample_data)
            return sample_data

        def generator_sample_caller():
            sample_times = count // chunksize
            for _ in range(sample_times):
                sample_data = self.model.sample(chunksize, **model_fit_kwargs)
                for d in self.data_processors:
                    sample_data = d.reverse_convert(sample_data)
                yield sample_data

            if count % chunksize > 0:
                sample_data = self.model.sample(count % chunksize, **model_fit_kwargs)
                for d in self.data_processors:
                    sample_data = d.reverse_convert(sample_data)
                yield sample_data

        return generator_sample_caller()

    def cleanup(self):
        if self.dataloader:
            self.dataloader.finalize(clear_cache=True)
        # Release resources
        del self.model

    def __del__(self):
        self.cleanup()
